from dataclasses import dataclass
from typing import Tuple


@dataclass
class VersionedType:
    name: str
    version: int

    def __key(self) -> Tuple[str, int]:
        return (self.name, self.version)

    def __hash__(self) -> int:
        return hash(self.__key)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, VersionedType):
            return self.__key == other.__key
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, VersionedType):
            if self.name == other.name:
                return self.version < other.version
            return self.name < other.name
        raise NotImplementedError


@dataclass
class EnrichedNamespace:
    name: str
    title: str

    def __key(self) -> Tuple[str, str]:
        return (self.name, self.title)

    def __hash__(self) -> int:
        return hash(self.__key)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, EnrichedNamespace):
            return self.__key == other.__key
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, EnrichedNamespace):
            if self.name == other.name:
                return self.title < other.title
            return self.name < other.name
        raise NotImplementedError



from typing import Union
from schema_registry.config import get_configuration
from schema_registry.registries.HttpRegistry import (
    EnrichingHttpSchemaRegistry,
    HttpSchemaRegistry,
)


def get_registry(
    profile: str = "default", extended: bool = False
) -> Union[EnrichingHttpSchemaRegistry, HttpSchemaRegistry]:
    config = get_configuration(profile)

    if extended:
        return EnrichingHttpSchemaRegistry(url=config.schema_registry_url)
    else:
        return HttpSchemaRegistry(url=config.schema_registry_url)
        # TODO: Unsupported methods



import logging
from typing import List
import typer

from rich.console import Console
from rich.logging import RichHandler

from schema_registry.app.ui.registry import app as reg_app

from schema_registry.app.business.cli import get_registry

console = Console()
app = typer.Typer()

FORMAT = "%(message)s"


@app.callback()
def main(
    ctx: typer.Context,
    verbose: List[bool] = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Output more details on what is happening. Use multiple times for more details.",
    ),
    profile: str = typer.Option(
        "default",
        "--profile",
        "-p",
        help="The profile to use for the command that is about to run.",
    ),
    extended: bool = typer.Option(
        False,
        "--extended",
        "-e",
        help="If enabled the output will be extended with explanatory text where available.",
    ),
) -> None:
    level = "WARNING"
    if len(verbose) == 1:
        level = "INFO"
    if len(verbose) >= 2:
        level = "DEBUG"

    logging.basicConfig(
        level=level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )

    ctx.obj = {"registry": get_registry(profile=profile, extended=extended)}


app.add_typer(reg_app, name="registry")

typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



from functools import lru_cache
import json
from typing import Any, Dict, List, Optional, Set

from requests_toolbelt.sessions import BaseUrlSession
from bs4 import BeautifulSoup
from bs4.element import ResultSet

from schema_registry.config import get_logger
from schema_registry.registries.entities import EnrichedNamespace, VersionedType


class HttpSchemaRegistry:
    def __init__(self, *, url: str) -> None:
        self.url = url
        self.session = BaseUrlSession(base_url=self.url)
        self.log = get_logger(self.__class__.__name__)
        self._namespaces: Set[str] = set()
        self._types: Dict[str, Set[VersionedType]] = {}

    @lru_cache(maxsize=128)
    def get(
        self, *, namespace: str, type_name: str, version: Optional[int] = None
    ) -> Optional[Dict[str, Any]]:
        if version is None:
            version = self.find_latest_version(namespace=namespace, type_name=type_name)

        req = self.session.get(
            f"/registry/{namespace}/{type_name}/{version}/schema.json"
        )

        self.log.debug(req)
        if req.status_code != 200:
            return None
        return req.json()  # type: ignore

    @property
    def namespaces(self) -> Set[str]:
        if not self._namespaces:
            self._initialize_namespaces()
        return self._namespaces

    def types(self, *, namespace: str) -> Set[VersionedType]:
        preliminary_result = self._types.get(namespace)

        if preliminary_result is None:
            self.log.debug(f"Fetching types for {namespace} from remote")
            self._initialize_types(namespace)

        return self._types.get(namespace, set())

    def _initialize_namespaces(self) -> None:
        index = self.session.get("/registry/index.json").json()
        schemes = index.get("schemes", [])

        self.log.debug(schemes)

        self._namespaces = {schema["ns"] for schema in schemes}

    def _initialize_types(self, namespace: str) -> None:
        index = self.session.get("/registry/index.json").json()
        schemes = index.get("schemes", [])
        filtered = filter(lambda x: x["ns"] == namespace, schemes)

        self._types[namespace] = {
            VersionedType(name=schema["type"], version=schema["version"])
            for schema in filtered
        }

    def find_latest_version(self, *, namespace: str, type_name: str) -> int:
        self.log.debug("Fetching version")
        types_of_namespace = self.types(namespace=namespace)

        found_types = list(filter(lambda x: x.name == type_name, types_of_namespace))
        found_types.sort()

        self.log.debug(f"Found {found_types}")

        # TODO: deal with case where found_types is empty/None
        version = found_types[-1].version

        if len(found_types) != 1:
            self.log.info(
                f"Found {len(found_types)} versions using latest version ({version})"
            )

        return version

    def refresh(self) -> None:
        self._initialize_namespaces()
        self._types = {}
        self.get.cache_clear()

    # TODO Transformation stuff


class EnrichingHttpSchemaRegistry:
    def __init__(self, *, url: str) -> None:
        self.url = url
        self.session = BaseUrlSession(base_url=self.url)
        self.http_schema_registry = HttpSchemaRegistry(url=url)
        self.log = get_logger(self.__class__.__name__)
        self._namespaces: Set[EnrichedNamespace] = set()
        self._types: Dict[str, Set[VersionedType]] = {}

    @lru_cache(maxsize=128)
    def get(
        self, *, namespace: str, type_name: str, version: Optional[int] = None
    ) -> Optional[Dict[str, Any]]:
        if version is None:
            version = self.http_schema_registry.find_latest_version(
                namespace=namespace, type_name=type_name
            )

        schema = None

        try:
            schema_code, _ = self._get_code_elements(
                namespace=namespace, type_name=type_name, version=version
            )  # type: ignore
            schema_description = self._get_schema_description(
                namespace=namespace, type_name=type_name
            )
            schema = json.loads(schema_code.get_text())
            schema["title"] = "\n".join(schema_description)

            self.log.debug(f"Found schema: {schema}")
        except TypeError:
            self.log.info("Requested schema not found.")
        except json.JSONDecodeError:
            self.log.warning(
                "Could not translate remote schema into JSON. HTML parsing or the page might be off."
            )

        return schema

    def _get_code_elements(
        self, *, namespace: str, type_name: str, version: int
    ) -> Optional[ResultSet]:

        # The way this is currently written is that 2 requests are made if the example and
        # the schema for the same event are requested. This is based on the assumption,
        # that users are interested in one or the other but not frequently in both.
        # Therefore trading of network traffic vs memory footprint.
        # * Should this prove false, we want lru_caching here as well, or maybe even
        # * a more powerful caching mechanism

        req = self.session.get(f"/{namespace}/{type_name.lower()}/{version}/index.html")

        if req.status_code != 200:
            return None

        raw_schema_html = req.content
        schema_html = BeautifulSoup(raw_schema_html, features="lxml")
        code_elements = schema_html.find_all("code")

        self.log.debug(f"Found {len(code_elements)} code elements")
        if len(code_elements) != 2:
            raise AttributeError("More than 2 code elements found can not parse schema")

        return code_elements

    def _get_schema_description(self, namespace: str, type_name: str) -> List[str]:
        req = self.session.get(f"/{namespace}/{type_name.lower()}/index.html")

        raw_schema_html = req.content
        schema_html = BeautifulSoup(raw_schema_html, features="lxml")
        headline = schema_html.body.h1

        description = []
        next_sibling = headline.next_sibling
        while next_sibling.name != "h2":
            self.log.debug(f"Sibling: {next_sibling}: {next_sibling.name}")
            if next_sibling.string and next_sibling.string.strip():
                description.append(next_sibling.string.strip())
            next_sibling = next_sibling.next_sibling

        return description

    @lru_cache(maxsize=128)
    def example(
        self, *, namespace: str, type_name: str, version: Optional[int] = None
    ) -> Optional[Dict[str, Any]]:
        if version is None:
            version = self.http_schema_registry.find_latest_version(
                namespace=namespace, type_name=type_name
            )

        example = None

        try:
            _, example_code = self._get_code_elements(
                namespace=namespace, type_name=type_name, version=version
            )  # type: ignore
            example = json.loads(example_code.get_text())

            self.log.debug(f"Found example: {example}")
        except TypeError:
            self.log.info("Requested example not found.")
        except json.JSONDecodeError:
            self.log.warning(
                "Could not translate remote example into JSON. HTML parsing or the page might be off."
            )

        return example

    @property
    def namespaces(self) -> Set[EnrichedNamespace]:
        if not self._namespaces:
            self._initialize_namespaces()
        return self._namespaces

    def _initialize_namespaces(self) -> None:
        namespaces = self.http_schema_registry.namespaces
        self._namespaces = set()
        for namespace in namespaces:
            req = self.session.get(f"/{namespace}")
            raw_schema_html = req.content
            namespace_html = BeautifulSoup(raw_schema_html, features="lxml")
            description = namespace_html.find("div", {"id": "body-inner"}).p

            if description and description.string:
                title = description.string
            else:
                title = ""

            self._namespaces.add(EnrichedNamespace(name=namespace, title=title))

    def types(self, *, namespace: str) -> Set[VersionedType]:
        return self.http_schema_registry.types(namespace=namespace)

    def refresh(self) -> None:
        self.http_schema_registry.refresh()
        self.get.cache_clear()

    # TODO Transformation stuff



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['schema_registry',
 'schema_registry.app',
 'schema_registry.app.business',
 'schema_registry.app.ui',
 'schema_registry.registries']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'jsonschema>=4.3.2,<5.0.0',
 'lxml>=4.7.1,<5.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'python-dotenv>=0.19,<0.20',
 'requests-toolbelt>=0.9,<0.10',
 'requests>=2,<3',
 'rich>=10,<11',
 'typer>=0,<1']

extras_require = \
{'docs': ['sphinx<4',
          'sphinx-click>=2.7.1,<3.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'sphinx-autodoc-typehints>=1.12.0,<2.0.0']}

entry_points = \
{'console_scripts': ['schemareg = schema_registry.app.ui.cli:app']}

setup_kwargs = {
    'name': 'schema-registry',
    'version': '0.0.6',
    'description': 'A schema registry implementation. Should you need to keep your json schemas in one place.',
    'long_description': '# Schema Registry\n\nJust a project stub for now\n\nContributions are welcome. Just get in touch.\n\n## Quickstart\n\nSimply `pip install schema-registry` and get going.\n\n## Development\n\nThis project uses `poetry` for dependency management and `pre-commit` for local checks.\n',
    'author': 'Eduard Thamm',
    'author_email': 'eduard.thamm@thammit.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/friendly-facts/schema-registry',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional
import typer

from rich.console import Console

app = typer.Typer()
console = Console()


@app.command()
def get_namespaces(ctx: typer.Context) -> None:
    res = list(ctx.obj["registry"].namespaces)
    res.sort()
    console.print(res)


@app.command()
def get_types(ctx: typer.Context, namespace: str) -> None:
    res = list(ctx.obj["registry"].types(namespace=namespace))
    res.sort()
    console.print(res)


@app.command()
def get_schema(
    ctx: typer.Context, namespace: str, type: str, version: Optional[int] = None
) -> None:
    console.print(
        ctx.obj["registry"].get(namespace=namespace, type_name=type, version=version)
    )


@app.command()
def get_example(
    ctx: typer.Context, namespace: str, type: str, version: Optional[int] = None
) -> None:
    console.print(
        ctx.obj["registry"].example(
            namespace=namespace, type_name=type, version=version
        )
    )
